<?php

namespace QUI\RabbitMQServer;

use PCSG\RabbitMQ\Channel\ChannelConfiguration;
use PCSG\Tasks\TaskDataEncryptionKey;
use PCSG\Tasks\TaskSerializer;
use QUI;
use QUI\Exception;
use QUI\Locale;
use QUI\QueueManager\Interfaces\QueueServerProviderInterface;
use QUI\RabbitMQServer\Interface\RabbitMQProviderInterface;

use function is_null;

class Provider implements QueueServerProviderInterface, RabbitMQProviderInterface
{
    private ?QUI\Config $serverConfig = null;

    public function getTitle(?Locale $locale = null): string
    {
        if (is_null($locale)) {
            $locale = QUI::getLocale();
        }

        return $locale->get('quiqqer/rabbitmqserver', 'server.title');
    }

    public function getDescription(?Locale $locale = null): string
    {
        if (is_null($locale)) {
            $locale = QUI::getLocale();
        }

        return $locale->get('quiqqer/rabbitmqserver', 'server.description');
    }

    /**
     * @throws Exception
     */
    public function getQueueServer(): Server
    {
        return new Server($this->getChannelConfiguration(), $this->getServerConfiguration());
    }

    public function getQueueServerClass(): string
    {
        return Server::class;
    }

    /**
     * @return TaskScheduler
     * @throws Exception
     */
    public function getTaskScheduler(): TaskScheduler
    {
        return new TaskScheduler(
            $this->getQueueServer(),
            $this->getTaskSerializer()
        );
    }

    /**
     * @return TaskSerializer
     * @throws Exception
     */
    public function getTaskSerializer(): TaskSerializer
    {
        return new TaskSerializer($this->getTaskDataEncryptionKey());
    }

    /**
     * @return TaskDataEncryptionKey|null
     * @throws Exception
     */
    public function getTaskDataEncryptionKey(): ?TaskDataEncryptionKey
    {
        $serverConfiguration = $this->getServerConfiguration();
        $taskDataEncryptionKey = $serverConfiguration->taskEncryptionKey;

        if (!$taskDataEncryptionKey) {
            if ($serverConfiguration->allowUnencryptedJobData) {
                return null;
            }

            // If unencrypted job data is not allowed -> use QUIQQER system salt as key
            QUI\System\Log::addWarning(
                "quiqqer/rabbitmqserver :: No task data encryption key configured AND unecrypted task data not"
                . " explicity permitted. Using QUIQQER system salt as fallback encryption key."
            );

            $taskDataEncryptionKey = QUI::conf('globals', 'salt');
        }

        return new TaskDataEncryptionKey($taskDataEncryptionKey);
    }

    /**
     * @return ChannelConfiguration
     * @throws Exception
     */
    private function getChannelConfiguration(): ChannelConfiguration
    {
        return new ChannelConfiguration(
            $this->getServerSetting('host'),
            $this->getServerSetting('port'),
            $this->getServerSetting('user'),
            $this->getServerSetting('password'),
            $this->getServerSetting('vhost') ?: '/',
            (bool)$this->getServerSetting('insist'),
            $this->getServerSetting('login_method') ?: 'AMQPLAIN',
            $this->getServerSetting('locale') ?: 'en_US',
            (float)$this->getServerSetting('connection_timeout') ?: 3.0,
            (float)$this->getServerSetting('read_write_timeout') ?: 3.0,
            (bool)$this->getServerSetting('keepalive'),
            (int)$this->getServerSetting('heartbeat') ?: 0
        );
    }

    /**
     * @return ServerConfiguration
     * @throws Exception
     */
    private function getServerConfiguration(): ServerConfiguration
    {
        return new ServerConfiguration(
            (bool)$this->getServerSetting('debug'),
            $this->getServerSetting('task_exchange') ?: null,
            $this->getServerSetting('task_data_encryption_key') ?: null,
            (bool)$this->getServerSetting('allow_unencrypted_job_data')
        );
    }

    /**
     * Get server setting
     *
     * @param string $key - config key
     * @return string
     * @throws Exception
     */
    private function getServerSetting(string $key): string
    {
        if (is_null($this->serverConfig)) {
            $this->serverConfig = QUI::getPackage('quiqqer/rabbitmqserver')->getConfig();
        }

        return $this->serverConfig->get('server', $key);
    }
}
