<?php

namespace QUI\RabbitMQServer;

use DateTime;
use PCSG\RabbitMQ\Channel\Channel as RabbitMQChannel;
use PCSG\RabbitMQ\Channel\ChannelConfiguration;
use PCSG\RabbitMQ\Channel\ChannelFactory;
use PCSG\RabbitMQ\Channel\ChannelFactoryInterface;
use PCSG\Tasks\TaskSerializerInterface;
use PCSG\Tasks\Utils;
use PCSG\Tasks\WorkerTask;
use PhpAmqpLib\Message\AMQPMessage;
use QUI;
use QUI\Exception;
use QUI\QueueManager\Interfaces\QueueJobRepositoryInterface;
use QUI\QueueManager\Interfaces\QueueServerInterface;
use QUI\QueueManager\JobData;
use QUI\QueueManager\JobLog;
use QUI\QueueManager\JobStatus;
use QUI\QueueManager\QueueJob;
use QUI\QueueManager\QueueJobUUID;
use QUI\QueueManager\QuiqqerWorkerQueueJob;
use QUI\QueueManager\Repository\QueueJobRepository;
use QUI\RabbitMQServer\Interface\TaskSchedulerInterface;
use QUI\Utils\System\File as FileUtils;

use function is_null;

/**
 * Class QueueServer
 *
 * RabbitMQ queue server class based on quiqqer/queuemanager
 *
 * @package QUI\QueueServer
 */
class Server implements QueueServerInterface
{
    private ?RabbitMQChannel $channel = null;

    /**
     * @param ChannelConfiguration $channelConfiguration
     * @param ServerConfiguration $serverConfiguration
     * @param ChannelFactoryInterface|null $channelFactory
     * @param QueueJobRepositoryInterface|null $queueJobRepository
     * @param TaskSerializerInterface|null $taskSerializer
     * @throws Exception
     */
    public function __construct(
        private readonly ChannelConfiguration $channelConfiguration,
        private readonly ServerConfiguration $serverConfiguration,
        private ?ChannelFactoryInterface $channelFactory = null,
        private ?QueueJobRepositoryInterface $queueJobRepository = null,
        private ?TaskSerializerInterface $taskSerializer = null
    ) {
        if (is_null($this->channelFactory)) {
            $this->channelFactory = new ChannelFactory();
        }

        if (is_null($this->queueJobRepository)) {
            $this->queueJobRepository = new QueueJobRepository();
        }

        if (is_null($this->taskSerializer)) {
            $provider = new Provider();
            $this->taskSerializer = $provider->getTaskSerializer();
        }
    }

    /**
     * Adds a single job to the queue of a server
     *
     * @param QueueJob $job - The job to add to the queue
     * @return QueueJobUUID - unique Job ID
     *
     * @throws QUI\Exception
     */
    public function queueJob(QueueJob $job): QueueJobUUID
    {
        if ($job instanceof QuiqqerWorkerQueueJob) {
            $jobUUID = $this->queueQuiqqerWorkerJob($job);
        } elseif ($this->isWorkerTaskQueueJob($job)) {
            $this->queueTaskJob($job);
            $jobUUID = $job->uuid;
        } else {
            throw new QUI\Exception(
                "quiqqer/rabbitmqserver can only queue jobs that are either of type"
                . " '" . QuiqqerWorkerQueueJob::class . "'"
                . " or that are specific QUIQQER-independent tasks scheduled with '"
                . TaskSchedulerInterface::class . "'"
            );
        }

        return $jobUUID;
    }

    /**
     * Check if a QueueJob is a special worker task job scheduled with the task scheduler.
     *
     * @param QueueJob $job
     * @return bool
     */
    private function isWorkerTaskQueueJob(QueueJob $job): bool
    {
        return $job->uuid && $job->data->get(JobDataKey::WORKER_TASK_SERIALIZED->value);
    }

    /**
     * @param QueueJob $job
     * @return void
     * @throws Exception
     */
    private function queueTaskJob(QueueJob $job): void
    {
        $this->pushDataToRabbitMQChannel(
            $job->data->get(JobDataKey::WORKER_TASK_SERIALIZED->value),
            $this->serverConfiguration->taskExchange,
            Utils::parseWorkerTaskDataClassToQueueName(
                $job->data->get(JobDataKey::WORKER_TASK_DATA_CLASS->value)
            ),
            $this->getDelayFromJob($job)
        );
    }

    /**
     * @param QuiqqerWorkerQueueJob $job
     * @return QueueJobUUID
     * @throws Exception
     */
    private function queueQuiqqerWorkerJob(QuiqqerWorkerQueueJob $job): QueueJobUUID
    {
        $job = $this->queueJobRepository->insert($job);

        // Only push job UUID to queue, since the consumer will fetch the job data from the database
        $this->pushDataToRabbitMQChannel($job->uuid, null, null, $this->getDelayFromJob($job));
        return $job->uuid;
    }

    /**
     * Pushes data to the RabbitMQ channel
     *
     * @param string $taskDataSerialized
     * @param string|null $exchange (optional) - If omitted use exchange name unique to this quiqqer system
     * @param string|null $queue (optional) - If omitted use queue name unique to this quiqqer system
     * @param int|null $delay (optional) - If omitted queue immediately
     *
     * @return void
     * @throws Exception
     * @throws \Exception
     */
    private function pushDataToRabbitMQChannel(
        string $taskDataSerialized,
        ?string $exchange = null,
        ?string $queue = null,
        ?int $delay = 0
    ): void {
        if (is_null($exchange)) {
            $exchange = $this->getUniqueQuiqqerExchangeName();
        }

        if (is_null($queue)) {
            $queue = $this->getUniqueQuiqqerQueueName();
        }

        $channel = $this->getChannel();

        if (!$channel->isExchangeDeclared($exchange)) {
            $channel->declareExchange(
                $exchange,
                [
                    'x-delayed-type' => 'direct'
                    // routing behaviour is identical to "direct", although exchange type is set for delayed messages
                ]
            );
        }

        if (!$channel->isQueueDeclared($queue)) {
            $channel->declareQueue(
                $queue,
                [
                    'x-max-priority' => ['I', 255]
                ]
            );
        }

        $channel->bindQueueToExchange($queue, $exchange, $queue);

        $channel->publish(
            $exchange,
            $queue,
            $taskDataSerialized,
            [
                'delivery_mode' => AMQPMessage::DELIVERY_MODE_PERSISTENT
            ],
            [
                'x-delay' => $delay
            ]
        );
    }

    /**
     * Get result of a specific job
     *
     * @param QueueJobUUID $jobId
     * @param bool $deleteJob (optional) - delete job from queue after return [default: true]
     * @return JobData
     *
     * @throws Exception
     */
    public function getJobResult(QueueJobUUID $jobId, bool $deleteJob = true): JobData
    {
        $job = $this->getJob($jobId);

        switch ($job->status) {
            case JobStatus::QUEUED:
                throw new QUI\Exception([
                    'quiqqer/queueserver',
                    'exception.queueserver.result.job.queued',
                    [
                        'jobId' => $jobId
                    ]
                ]);

            case JobStatus::RUNNING:
                throw new QUI\Exception([
                    'quiqqer/queueserver',
                    'exception.queueserver.result.job.running',
                    [
                        'jobId' => $jobId
                    ]
                ]);
        }

        if ($deleteJob) {
            $this->deleteJob($jobId);
        }

        return $job->result;
    }


    /**
     * Set result of a specific job
     *
     * @param QueueJobUUID $jobId
     * @param JobData $result
     * @return void
     * @throws Exception
     */
    public function setJobResult(QueueJobUUID $jobId, JobData $result): void
    {
        $job = $this->getJob($jobId);

        Server::writeDebugMsg('Setting job result Job #' . $jobId);

        switch ($job->status) {
            case JobStatus::FINISHED:
                throw new QUI\Exception(
                    "Cannot set job result for $jobId :: Job is already finished."
                );

            case JobStatus::ERROR:
                throw new QUI\Exception(
                    "Cannot set job result for $jobId :: Job is in error state."
                );
        }

        $job->result = $result;
        $this->queueJobRepository->update($job);
    }

    /**
     * Set status of a job
     *
     * @param QueueJobUUID $jobId
     * @param JobStatus $status
     * @return void
     * @throws Exception
     */
    public function setJobStatus(QueueJobUUID $jobId, JobStatus $status): void
    {
        $job = $this->getJob($jobId);

        if ($job->status === $status) {
            return;
        }

        $job->status = $status;
        $this->queueJobRepository->update($job);
    }

    /**
     * Get status of a job
     *
     * @param QueueJobUUID $jobId
     * @return JobStatus
     * @throws Exception
     */
    public function getJobStatus(QueueJobUUID $jobId): JobStatus
    {
        $job = $this->getJob($jobId);
        return $job->status;
    }

    /**
     * Write log entry for a job
     *
     * @param QueueJobUUID $jobId
     * @param string $msg
     * @return void
     * @throws Exception
     */
    public function writeJobLogEntry(QueueJobUUID $jobId, string $msg): void
    {
        $job = $this->getJob($jobId);

        $job->log->addEntry([
            'time' => date('Y.m.d H:i:s'),
            'msg' => $msg
        ]);

        $this->queueJobRepository->update($job);
    }

    /**
     * Get event log for specific job
     *
     * @param QueueJobUUID $jobId
     * @return JobLog
     * @throws Exception
     */
    public function getJobLog(QueueJobUUID $jobId): JobLog
    {
        $job = $this->getJob($jobId);
        return $job->log;
    }

    /**
     * Delete a job
     *
     * @param QueueJobUUID $jobId
     * @return void
     */
    public function deleteJob(QueueJobUUID $jobId): void
    {
        // currently not possible with RabbitMQ
        // https://discuss.pivotal.io/hc/en-us/community/posts/205630238-How-to-delete-messages-with-an-specific-routing-key-from-a-queue-
    }

    /**
     * Clone a job and queue it immediately
     *
     * @param QueueJobUUID $jobId - Job UUID
     * @param integer $priority - (new) job priority
     * @param DateTime|null $earliestQueueDate (optional) - Date at which the job shall be queued/executed the earliest
     *
     * @return QueueJobUUID - ID of cloned job
     * @throws Exception
     */
    public function cloneJob(QueueJobUUID $jobId, int $priority, ?DateTime $earliestQueueDate = null): QueueJobUUID
    {
        $currentJob = $this->getJob($jobId);

        if ($currentJob instanceof QuiqqerWorkerQueueJob) {
            $clonedJob = new QuiqqerWorkerQueueJob(
                $currentJob->getQuiqqerQueueWorkerClass(),
                null,
                $currentJob->jobAttributes
            );

            $clonedJob->data->setAll($currentJob->data->getAll());
        } elseif ($this->isWorkerTaskQueueJob($currentJob)) {
            $cloneJobUuid = new QueueJobUUID(QUI\Utils\Uuid::get());

            $clonedJob = new QueueJob(
                $cloneJobUuid,
                $currentJob->jobAttributes
            );

            $clonedJob->data->setAll($currentJob->data->getAll());

            $workerTask = $this->taskSerializer->deserializeWorkerTaskFromQueue(
                $currentJob->data->get(JobDataKey::WORKER_TASK_SERIALIZED->value)
            );

            $clonedWorkerTask = new WorkerTask(
                $cloneJobUuid,
                $workerTask->taskData,
                $workerTask->logLevel,
                $workerTask->earliestDeliveryDate
            );

            $clonedJob->data->set(
                JobDataKey::WORKER_TASK_SERIALIZED->value,
                $this->taskSerializer->serializeWorkerTaskForQueue($clonedWorkerTask)
            );

            $clonedJob->data->set(
                JobDataKey::WORKER_TASK_DATA_CLASS->value,
                $workerTask->taskData::class
            );
        } else {
            throw new QUI\Exception(
                "quiqqer/rabbitmqserver can only clone jobs that are either of type"
                . " '" . QuiqqerWorkerQueueJob::class . "'"
                . " or that are specific QUIQQER-independent tasks scheduled with '"
                . TaskSchedulerInterface::class . "'"
            );
        }

        $clonedJob->jobAttributes->earliestQueueDate = $earliestQueueDate;
        $clonedJob->jobAttributes->priority = $priority;

        $clonedJobUuid = $this->queueJob($clonedJob);

        // Set "CLONED" status to cloned job
        $currentJob->status = JobStatus::CLONED;

        $clonedJobUuids = $currentJob->data->get(JobDataKey::CLONED_JOB_UUIDS->value);

        if (empty($clonedJobUuids)) {
            $clonedJobUuids = [];
        }

        $clonedJobUuids[] = $clonedJobUuid;
        $currentJob->data->set(JobDataKey::CLONED_JOB_UUIDS->value, $clonedJobUuids);
        $this->queueJobRepository->update($currentJob);

        return $clonedJobUuid;
    }

    /**
     * Check if a connection to a RabbitMQ server is currently established
     *
     * @return bool
     */
    public function isConnected(): bool
    {
        if (is_null($this->channel)) {
            return false;
        }

        return $this->channel->isConnected();
    }

    /**
     * Close server connection
     */
    public function closeConnection(): void
    {
        $this->channel?->closeConnection();
    }

    /**
     * Get unique queue name for this quiqqer system
     *
     * @return string
     * @throws Exception
     */
    public function getUniqueQuiqqerQueueName(): string
    {
        $varDir = QUI::getPackage('quiqqer/rabbitmqserver')->getVarDir();
        $queueNameFile = $varDir . 'queuename';
        $queueName = FileUtils::getFileContent($queueNameFile);

        if (!empty($queueName)) {
            return $queueName;
        }

        // if queue name has not been generated yet -> generate
        $queueName = uniqid('quiqqer_queue_', true);
        FileUtils::mkfile($queueNameFile);
        file_put_contents($queueNameFile, $queueName);

        return $queueName;
    }

    /**
     * Get unique RabbitMQ exchange name for this quiqqer system
     *
     * @return string
     * @throws Exception
     */
    private function getUniqueQuiqqerExchangeName(): string
    {
        $varDir = QUI::getPackage('quiqqer/rabbitmqserver')->getVarDir();
        $exchangeNameFile = $varDir . 'exchange';
        $exchangeName = FileUtils::getFileContent($exchangeNameFile);

        if (!empty($exchangeName)) {
            return $exchangeName;
        }

        // if name has not been generated yet -> generate
        $exchangeName = uniqid('quiqqer_exchange_', true);
        FileUtils::mkfile($exchangeNameFile);
        file_put_contents($exchangeNameFile, $exchangeName);

        return $exchangeName;
    }

    /**
     * Write debug msg in special log file.
     *
     * Requires "debug" setting to be 1.
     *
     * @param string $msg
     * @return void
     */
    public function writeDebugMsg(string $msg): void
    {
        if (!$this->serverConfiguration->debug) {
            return;
        }

        QUI\System\Log::write(
            $msg,
            QUI\System\Log::LEVEL_INFO,
            [],
            'rabbit_debug'
        );
    }

    /**
     * @param bool $forceReconnect (optional) - force new channel
     *
     * @return RabbitMQChannel
     * @throws \Exception
     */
    public function getChannel(bool $forceReconnect = false): RabbitMQChannel
    {
        if (is_null($this->channel) || !$this->isConnected() || $forceReconnect) {
            $this->closeConnection();
            $this->channel = $this->channelFactory->buildChannel($this->channelConfiguration);
        }

        return $this->channel;
    }

    /**
     * @param QueueJobUUID $jobId
     * @return QueueJob
     * @throws Exception
     */
    private function getJob(QueueJobUUID $jobId): QueueJob
    {
        $job = $this->queueJobRepository->find($jobId);

        if (is_null($job)) {
            throw new QUI\Exception("Queue job $jobId not found.");
        }

        return $job;
    }

    /**
     * Calculate message delay of a job.
     *
     * @param QueueJob $job
     * @return int - Delay in seconds
     */
    private function getDelayFromJob(QueueJob $job): int
    {
        if (is_null($job->getEarliestQueueDate())) {
            return 0;
        }

        return Utils::calculateMessageDelayFromDate($job->getEarliestQueueDate());
    }
}
